# -*- coding: utf-8 -*-

# BASED ON https://github.com/apache/airflow/blob/349b0811c3022605426ba57d30936240a7c2848a/airflow/providers/snowflake/hooks/snowflake.py
# Original code license
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import logging
import sys

from contextlib import closing
from typing import Any, Iterable, List, Mapping, Tuple, Union

import six

from airflow.contrib.hooks.snowflake_hook import SnowflakeHook


logger = logging.getLogger(__name__)


def snowflake_get_records(
    hook: SnowflakeHook, sql: str, parameters: Union[Mapping, Iterable, None] = None
) -> Tuple[List, int, str]:
    """
    Executes the sql and returns a set of records alongside with
    session_id and query_id generated by Snowflake.

    :param hook: Configured SnowflakeHook with Snoflake connection details.
    :param sql: the sql statement to be executed (str) or a list of
        sql statements to execute
    :param parameters: The parameters to render the SQL query with.
    :return: Query result alongside with session_id and query_id
        assigned by Snowflake to each submitted query
    """
    if sys.version_info[0] < 3:
        sql = sql.encode("utf-8")

    with closing(hook.get_conn()) as conn:
        session_id = conn.session_id
        with closing(conn.cursor()) as cur:
            if parameters is not None:
                cur.execute(sql, parameters)
            else:
                cur.execute(sql)
            query_id = cur.sfqid
            return cur.fetchall(), session_id, query_id


def snowflake_get_first(
    hook: SnowflakeHook, sql: str, parameters: Union[Mapping, Iterable, None] = None
) -> Tuple[Any, int, str]:
    """
    Executes the sql and returns the first resulting row alongside with
    session_id and query_id generated by Snowflake.

    :param hook: Configured SnowflakeHook with Snoflake connection details.
    :param sql: the sql statement to be executed (str) or a list of
        sql statements to execute
    :param parameters: The parameters to render the SQL query with.
    :return: Query result alongside with session_id and query_id
        assigned by Snowflake to each submitted query
    """
    if sys.version_info[0] < 3:
        sql = sql.encode("utf-8")

    with closing(hook.get_conn()) as conn:
        session_id = conn.session_id
        with closing(conn.cursor()) as cur:
            if parameters is not None:
                cur.execute(sql, parameters)
            else:
                cur.execute(sql)
            query_id = cur.sfqid
            return cur.fetchone(), session_id, query_id


def snowflake_run(
    hook: SnowflakeHook,
    sql: Union[str, List[str]],
    autocommit: bool = False,
    parameters: Union[Mapping, Iterable, None] = None,
) -> Tuple[int, List[str]]:
    """
    Runs a command or a list of commands and returns session_id and query_id(s).
    Pass a list of sql statements to the sql parameter to get them to execute sequentially.

    :param hook: Configured SnowflakeHook with Snoflake connection details.
    :param sql: the sql statement to be executed (str) or a list of
        sql statements to execute
    :param autocommit: What to set the connection's autocommit setting to
        before executing the query.
    :param parameters: The parameters to render the SQL query with.
    :return: session_id and query_id(s) assigned by Snowflake to each submitted query
    """
    if isinstance(sql, six.string_types):
        sql = [sql]

    with closing(hook.get_conn()) as conn:
        session_id, query_ids = conn.session_id, []
        if hook.supports_autocommit:
            hook.set_autocommit(conn, autocommit)

        with closing(conn.cursor()) as cur:
            for s in sql:
                if sys.version_info[0] < 3:
                    s = s.encode("utf-8")
                if parameters is not None:
                    hook.log.info("{} with parameters {}".format(s, parameters))
                    res = cur.execute(s, parameters)
                else:
                    hook.log.info(s)
                    res = cur.execute(s)
                query_ids.append(res.sfqid)

        # If autocommit was set to False for db that supports autocommit,
        # or if db does not supports autocommit, we do a manual commit.
        if not hook.get_autocommit(conn):
            conn.commit()
    logger.info(
        "Executed queries '{}', got session_id {}, query_id {}".format(
            sql, session_id, query_ids
        )
    )
    return (session_id, query_ids)


class DbndSnowflakeHook(SnowflakeHook):
    """
    Use this Hook instead of `airflow.contrib.hooks.snowflake_hook.SnowflakeHook` if you
    need session_id and query_id(s) generated by Snowflake.
    E.g. to pass them to `dbnd_snowflake.snowflake_resources.log_snowflake_resource_usage()`.
    """

    @staticmethod
    def from_snowflake_hook(hook: SnowflakeHook) -> "DbndSnowflakeHook":
        return DbndSnowflakeHook(
            snowflake_conn_id=getattr(hook, hook.conn_name_attr),
            warehouse=hook.warehouse,
            account=hook.account,
            database=hook.database,
            region=hook.region,
            role=hook.role,
            schema=hook.schema,
        )

    def get_records(
        self, sql: str, parameters: Union[Mapping, Iterable, None] = None
    ) -> Tuple[List, int, str]:
        """
        Executes the sql and returns a set of records alongside with
        session_id and query_id generated by Snowflake.

        :param sql: the sql statement to be executed (str) or a list of
            sql statements to execute
        :param parameters: The parameters to render the SQL query with.
        :return: Query result alongside with session_id and query_id
            assigned by Snowflake to each submitted query
        """
        return snowflake_get_records(self, sql, parameters)

    def get_first(
        self, sql: str, parameters: Union[Mapping, Iterable, None] = None
    ) -> Tuple[Any, int, str]:
        """
        Executes the sql and returns the first resulting row alongside with
        session_id and query_id generated by Snowflake.

        :param sql: the sql statement to be executed (str) or a list of
            sql statements to execute
        :param parameters: The parameters to render the SQL query with.
        :return: Query result alongside with session_id and query_id
            assigned by Snowflake to each submitted query
        """
        return snowflake_get_first(self, sql, parameters)

    def run(
        self,
        sql: Union[str, List[str]],
        autocommit: bool = False,
        parameters: Union[Mapping, Iterable, None] = None,
    ) -> Tuple[int, List[str]]:
        """
        Runs a command or a list of commands. Pass a list of sql
        statements to the sql parameter to get them to execute
        sequentially

        :param sql: the sql statement to be executed (str) or a list of
            sql statements to execute
        :param autocommit: What to set the connection's autocommit setting to
            before executing the query.
        :param parameters: The parameters to render the SQL query with.
        :return: session_id and query_id(s) assigned by Snowflake to each submitted query
        """
        return snowflake_run(self, sql, autocommit, parameters)



from typing import Optional

from dbnd import log_duration
from dbnd._core.commands.metrics import log_data
from dbnd._core.plugin.dbnd_plugins import is_plugin_enabled
from dbnd_snowflake.snowflake_config import SnowflakeConfig


def log_snowflake_table(
    table_name: str,
    connection_string: str,
    database: str,
    schema: str,
    key: Optional[str] = None,
    with_preview: Optional[bool] = None,
    with_schema: Optional[bool] = None,
    raise_on_error: bool = False,
):

    if not is_plugin_enabled("dbnd-snowflake", module_import="dbnd_snowflake"):
        return
    from dbnd_snowflake import snowflake_values

    with log_duration("log_snowflake_table__time_seconds", source="system"):
        conn_params = snowflake_values.conn_str_to_conn_params(connection_string)
        account = conn_params["account"]
        user = conn_params["user"]
        password = conn_params["password"]

        config = SnowflakeConfig()
        snowflake_table = snowflake_values.SnowflakeTable(
            account,
            user,
            password,
            database,
            schema,
            table_name,
            config.table_preview_rows,
        )
        log_data(
            key or table_name,
            snowflake_table,
            with_preview=with_preview,
            with_schema=with_schema,
            with_size=with_schema,
            with_histograms=False,
            raise_on_error=raise_on_error,
        )



import logging

from textwrap import dedent
from typing import Dict

import attr
import snowflake.connector

from six.moves.urllib_parse import urlparse
from snowflake.connector import DictCursor

from dbnd import __version__
from dbnd._core.errors import DatabandConfigError, DatabandRuntimeError
from dbnd._core.utils.string_utils import humanize_bytes
from dbnd._vendor.tabulate import tabulate
from targets.value_meta import ValueMeta, ValueMetaConf
from targets.values import register_value_type
from targets.values.builtins_values import DataValueType


logger = logging.getLogger(__name__)


class SnowflakeError(DatabandRuntimeError):
    pass


@attr.s
class SnowflakeTable(object):
    account = attr.ib()  # type: str
    user = attr.ib()  # type: str
    password = attr.ib()  # type: str
    database = attr.ib()  # type: str
    schema = attr.ib()  # type: str
    table_name = attr.ib()  # type: str
    preview_rows = attr.ib(default=20)  # type: int


class SnowflakeTableValueType(DataValueType):
    type = SnowflakeTable
    type_str = "SnowflakeTable"
    support_merge = False

    config_name = "snowflake_table"

    def get_snowflake(self, value):
        # TODO: Use sqlalchemy.engine.url here
        conn_string = SnowflakeController.conn_params_to_conn_str(
            account=value.account,
            user=value.user,
            password=value.password,
            database=value.database,
            schema=value.schema,
        )
        return SnowflakeController(conn_string)

    def to_signature(self, x):
        # type: (SnowflakeTable) -> str
        # don't include user and password in uri
        return "snowflake://{0.user}:***@{0.account}/{0.database}.{0.schema}/{0.table_name}".format(
            x
        )

    def get_value_meta(self, value, meta_conf):
        # type: (SnowflakeTable, ValueMetaConf) -> ValueMeta
        data_schema = {}
        data_preview = data_dimensions = None

        with self.get_snowflake(value) as snowflake:
            stats, histograms = {}, {}
            hist_sys_metrics = None
            if meta_conf.log_preview:
                data_preview = snowflake.to_preview(value)
            if meta_conf.log_schema:
                data_schema = {
                    "type": self.type_str,
                    "column_types": snowflake.get_column_types(value),
                }
            if meta_conf.log_size:
                dimensions = snowflake.get_dimensions(value)
                data_dimensions = [dimensions["rows"], dimensions["cols"]]
                data_schema["size"] = humanize_bytes(dimensions["bytes"])

        return ValueMeta(
            value_preview=data_preview,
            data_dimensions=data_dimensions,
            data_schema=data_schema,
            data_hash=self.to_signature(value),
            descriptive_stats=stats,
            histogram_system_metrics=hist_sys_metrics,
            histograms=histograms,
        )


register_value_type(SnowflakeTableValueType())

# TODO? Subclass from SingletonContext to reuse connection??
class SnowflakeController:
    """ Interacts with Snowflake, queries it"""

    def __init__(self, connection_string):
        conn_params = conn_str_to_conn_params(connection_string)

        self.connection_string = connection_string
        self.account = conn_params["account"]
        self.user = conn_params["user"]
        self.password = conn_params["password"]

        self._query_tag = "dbnd-snowflake"
        self._connection = None
        self._column_types = None

    @staticmethod
    def conn_params_to_conn_str(**conn_params):
        # type: (**Dict) -> str
        # TODO: Drop in favor of sqlalchemy.engine.url
        uri = "snowflake://{user}:{password}@{account}/"
        return uri.format(**conn_params)

    def __enter__(self):
        self._connection = snowflake.connector.connect(
            user=self.user,
            password=self.password,
            account=self.account,
            session_parameters={"QUERY_TAG": self._query_tag},
            application="DBND Snowflake plugin {}".format(__version__),
        )
        self._cursor = self._connection.cursor(DictCursor)

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._connection is not None:
            self._connection.close()

    def to_preview(self, table):
        # type: (SnowflakeTable) -> str
        # snowflake.connector does not handle incorrect utf-8 data fetched from db,
        # hence this fillding with encode/decode
        column_types = self.get_column_types(table)
        columns = ",".join(
            'TRY_HEX_DECODE_STRING(HEX_ENCODE("{0}")) AS {0}'.format(column)
            for column in column_types.keys()
        )

        rows = self._query(
            "select {0} from {1.database}.{1.schema}.{1.table_name} limit {1.preview_rows}".format(
                columns, table
            )
        )
        preview_table = tabulate(rows, headers="keys") + "\n..."
        return preview_table

    def get_dimensions(self, table):
        # type: (SnowflakeTable) -> Dict
        table_meta = self._query(
            "SHOW TABLES LIKE '{0.table_name}' in schema {0.database}.{0.schema}".format(
                table
            )
        )
        if len(table_meta) != 1:
            raise SnowflakeError(
                "Snowflake table not found: '{0.table_name}', DB: '{0.database}'".format(
                    table
                )
            )

        cols = len(self.get_column_types(table))
        return {
            "rows": table_meta[0]["rows"],
            "cols": cols,
            "bytes": table_meta[0]["bytes"],
        }

    def get_column_types(self, table):
        # type: (SnowflakeTable) -> Dict[str, str]
        if self._column_types is not None:
            return self._column_types

        query = dedent(
            """\
            SELECT column_name, data_type
            FROM {0.database}.information_schema.columns
            WHERE LOWER(table_name) = LOWER('{0.table_name}')
                and LOWER(table_schema) = LOWER('{0.schema}')"""
        ).format(table)
        results = self._query(query)
        if not results:
            raise SnowflakeError(
                "Table columns not found. Snowflake DB: '{0.database}', "
                "schema: {0.schema} table: '{0.table_name}'\n"
                "Query used: {1}".format(table, query)
            )

        self._column_types = {row["COLUMN_NAME"]: row["DATA_TYPE"] for row in results}
        return self._column_types

    def _query(self, query, params=None):
        try:
            self._cursor.execute(query, params)
            result = self._cursor.fetchall()
            return result
        except Exception:
            logger.exception(
                "Error occurred during querying Snowflake, query: %s", query
            )
            raise


def conn_str_to_conn_params(conn_str):
    # type: (str) -> dict
    # TODO: Drop in favor of sqlalchemy.engine.url
    conn = urlparse(conn_str)
    if conn.scheme != "snowflake":
        raise DatabandConfigError(
            "Unsupported connection string scheme '{}'. snowflake is required".format(
                conn.scheme
            )
        )

    return {
        "account": conn.hostname,
        "user": conn.username,
        "password": conn.password,
    }



from dbnd_snowflake.snowflake_resources import log_snowflake_resource_usage
from dbnd_snowflake.snowflake_table import log_snowflake_table


__all__ = ["log_snowflake_resource_usage", "log_snowflake_table"]



import logging

from datetime import timedelta
from decimal import Decimal
from textwrap import dedent
from time import sleep
from typing import Dict, List, Optional, Tuple

from dbnd import log_duration, log_metrics
from dbnd._core.utils.timezone import utcnow
from dbnd_snowflake.snowflake_config import SnowflakeConfig
from dbnd_snowflake.snowflake_values import SnowflakeController, SnowflakeError


logger = logging.getLogger(__name__)

SNOWFLAKE_METRIC_TO_UI_NAME = {
    "BYTES_SCANNED": "bytes_scanned",
    "COMPILATION_TIME": "compilation_time_milliseconds",
    "CREDITS_USED_CLOUD_SERVICES": "credits_used_cloud_services",
    "ERROR_MESSAGE": "error_message",
    "EXECUTION_STATUS": "execution_status",
    "EXECUTION_TIME": "execution_time_milliseconds",
    "QUERY_ID": "query_id",
    "QUERY_TAG": "query_tag",
    "QUERY_TEXT": "query_text",
    "ROWS_PRODUCED": "rows_produced",
    "SESSION_ID": "session_id",
    "TOTAL_ELAPSED_TIME": "total_elapsed_time_milliseconds",
}
RESOURCE_METRICS = ",".join(
    '"{}"'.format(m) for m in SNOWFLAKE_METRIC_TO_UI_NAME.keys()
)
RESULT_LIMIT_INC = 10


def log_snowflake_resource_usage(
    database: str,
    connection_string: str,
    query_ids: List[str],
    session_id: Optional[int] = None,
    key: str = "snowflake_query",
    history_window: float = 15,
    query_history_result_limit: Optional[int] = None,
    delay: int = 0,
    retries: int = 3,
    retry_pause: float = 0,
    raise_on_error: bool = False,
) -> None:
    """
    Search for a query previously executed by Snowflake in it's QUERY_HISTORY and log cpu time,
    run time, disk read, and other resources.

    Query's metadata can appear in QUERY_HISTORY with a lag up to 45 minutes.

    :param database: Name of the database query was issued to.
    :param connection_string: Snowflake connection string to use.
    :param query_ids: Supply a list of `query_id` generated by Snowflake for search in QUERY_HISTORY.
    :param session_id: Supply `session_id` generated by Snowflake for more efficient search in QUERY_HISTORY.
    :param key: Override it if you call this function twice or more within the same task/Airflow Operator
    :param history_window: How deep to search into QUERY_HISTORY. Set in minutes
    :param query_history_result_limit: Passed through directly to QUERY_HISTORY search function as `RESULT_LIMIT` param
    :param delay: Initial delay before looking in QUERY_HISTORY.
        Metadata can appear there with some delay. Use this param for fine tuning
    :param retries: How much times to search in QUERY_HISTORY.
        Each time search is widened by increasing `RESULT_LIMIT` param.
    :param raise_on_error: By default all exceptions are muted so your task success status
        is not affected by errors in tracking. Set to true to re-raise all exceptions.
    :param retry_pause: Set number of seconds to pause before next retry.
    """

    snowflake_config = SnowflakeConfig()
    if query_history_result_limit is None:
        query_history_result_limit = snowflake_config.query_history_result_limit

    if not all(query_id for query_id in query_ids):
        error_msg = f"query_ids cannot be empty. You supplied: {query_ids}"
        if raise_on_error:
            raise SnowflakeError(error_msg)
        else:
            logger.error(error_msg)
            return

    metrics_to_log = {}
    # XXX: Do we actually need log_duration?
    with log_duration("log_snowflake_resource_usage__time_seconds", "system"):
        for i, query_id in enumerate(query_ids):
            query_key = f"{key}.{i}" if len(query_ids) > 1 else key
            metrics_to_log.update(
                _get_snowflake_resource_usage(
                    database,
                    connection_string,
                    query_id,
                    session_id,
                    query_key,
                    history_window,
                    query_history_result_limit,
                    delay,
                    retries,
                    retry_pause,
                    raise_on_error,
                    snowflake_config,
                )
            )

    log_metrics(metrics_to_log, source="user")


def _get_snowflake_resource_usage(
    database: str,
    connection_string: str,
    query_id: str,
    session_id: Optional[int],
    key: str,
    history_window: float,
    query_history_result_limit: int,
    delay: int,
    retries: int,
    retry_pause: float,
    raise_on_error: bool,
    config: SnowflakeConfig,
) -> Dict:
    if delay > 0:
        logger.info("Delaying search in QUERY_HISTORY for %s seconds", delay)
        sleep(delay)
    result_limit = min(
        query_history_result_limit, config.query_history_result_limit_max_value
    )
    tries, sf_query = 0, ""
    try:
        while (
            result_limit <= config.query_history_result_limit_max_value
            and tries <= retries
        ):
            resource_metrics, sf_query = _query_snowflake_resource_usage(
                database,
                connection_string,
                query_id=query_id,
                session_id=session_id,
                key=key,
                history_window=history_window,
                query_history_result_limit=result_limit,
                config=config,
            )
            if resource_metrics:
                return resource_metrics
            logger.warning(
                "Metadata not found for session_id '{}', query_id '{}'\n"
                "Query used to search for resource usage: '{}'".format(
                    session_id, query_id, sf_query
                )
            )
            result_limit = min(
                result_limit * RESULT_LIMIT_INC,
                config.query_history_result_limit_max_value,
            )
            logger.info(
                "Extending QUERY_HISTORY() search window: RESULT_LIMIT={}".format(
                    result_limit
                )
            )
            tries += 1
            if retry_pause and retry_pause > 0:
                logger.info("Sleeping for %s seconds", retry_pause)
                sleep(retry_pause)
        else:
            logger.info(
                "No more retries left to fetch Snoflake query resources. Giving up."
            )

    except Exception as exc:
        conn_without_pass = _censor_password(connection_string)
        logger.exception(
            "Failed to log_snowflake_resource_usage (query_text=%s, connection_string=%s)\n"
            "Last query params used to search for resource usage: query_id - '%s', "
            "sesion_id = '%s', database - '%s', connection - '%s', query - '%s'",
            query_id,
            session_id,
            database,
            conn_without_pass,
            sf_query,
        )
        if raise_on_error:
            raise

    logger.error(
        "Resource metrics were not found for query_id '%s'.\n Query used: %s",
        query_id,
        sf_query,
    )
    return {
        f"{key}.warning": "No resources info found",
        # converting to str, since can be too large for DB int
        f"{key}.session_id": str(session_id),
        f"{key}.query_id": query_id,
    }


def _build_snowflake_resource_usage_query(
    database: str,
    query_id: str,
    session_id: int,
    history_window: float,
    query_history_result_limit: int,
    config: SnowflakeConfig,
) -> str:

    time_end = utcnow() - timedelta(minutes=config.query_history_end_time_range_end)
    time_start = time_end - timedelta(
        minutes=history_window or config.query_history_end_time_range_start
    )
    if session_id:
        query_history = dedent(
            """\
            select {metrics}
            from table({database}.information_schema.query_history_by_session(
                SESSION_ID => {session_id},
                END_TIME_RANGE_START => '{time_start}'::timestamp_ltz,
                END_TIME_RANGE_END => '{time_end}'::timestamp_ltz,
                RESULT_LIMIT => {result_limit}
            ))
            where query_id='{query_id}'
            order by start_time desc limit 1;"""
        ).format(
            metrics=RESOURCE_METRICS,
            database=database,
            minutes=history_window,
            session_id=session_id,
            result_limit=query_history_result_limit,
            time_start=time_start,
            time_end=time_end,
            query_id=query_id,
        )
        return query_history

    else:
        query_history = dedent(
            """\
            select {metrics}
            from table({database}.information_schema.query_history(
                END_TIME_RANGE_START => '{time_start}'::timestamp_ltz,
                END_TIME_RANGE_END => '{time_end}'::timestamp_ltz,
                RESULT_LIMIT => {result_limit}
            ))
            where query_id='{query_id}'
            order by start_time desc limit 1;"""
        ).format(
            metrics=RESOURCE_METRICS,
            database=database,
            minutes=history_window,
            result_limit=query_history_result_limit,
            time_start=time_start,
            time_end=time_end,
            query_id=query_id,
        )
        return query_history


def _query_snowflake_resource_usage(
    database,  # type: str
    connection_string,  # type: str
    query_id,  # type: str
    session_id,  # type: Optional[int]
    key,  # type: Optional[str]
    history_window,  # type: float
    query_history_result_limit,  # type: int
    config,  # type: SnowflakeConfig
):  # type: (...) -> Tuple[Dict, str]
    key = key or "snowflake_query"
    query_history = _build_snowflake_resource_usage_query(
        database,
        query_id=query_id,
        session_id=session_id,
        history_window=history_window,
        query_history_result_limit=query_history_result_limit,
        config=config,
    )

    result = _connect_and_query(connection_string, query_history)
    if not result:
        return {}, query_history

    metrics = result[0]

    metrics_to_log = {}
    for metric, ui_name in SNOWFLAKE_METRIC_TO_UI_NAME.items():
        if metric in metrics:
            value = metrics[metric]
            # Quick hack to track decimal values. probably should be handled on a serialization level
            if isinstance(value, Decimal):
                value = float(value)
            metrics_to_log[key + "." + ui_name] = value
    return metrics_to_log, query_history


def _connect_and_query(connection_string, query, *params):
    """ connect if needed, then query. """
    if not connection_string:
        logger.error("Connection string is empty, don't know where to connect")
        return

    with SnowflakeController(connection_string) as snowflake:
        return snowflake._query(query, params)


def _censor_password(connection_string):
    """
    example connection string:
        snowflake://user:password@account.europe-west4.gcp/database
    returns:
        snowflake://user:*****@account.europe-west4.gcp/database
    """
    if (not connection_string) or ("@" not in connection_string):
        return connection_string

    split1 = connection_string.split("@")
    split2 = split1[0].split(":")

    if len(split2) != 3:
        return connection_string

    split2[-1] = "*****"
    split2_join = ":".join(split2)
    split1[0] = split2_join
    split1_join = "@".join(split1)
    return split1_join
